/*
* Nebulizer Player Pro v1.8
* Программа-анализатор спектра
* для управления ЦМУ Nebulizer Pro
* Разработка: Панкратьев Д. И., 2020
*/

import ddf.minim.*;
import ddf.minim.analysis.*;
import processing.serial.*;
import hypermedia.net.*; //UDP library
import g4p_controls.*;
import java.util.Arrays;

Minim minim;
AudioPlayer player;
FFT fft;
BeatDetect beat;
BeatListener bl;
AudioMetaData meta;

Serial port;
File f;
UDP udp;

GKnob kb0, kb1, kb2, kb3, kb4, kb5, kb6, kb7, kb8;
GWindow window;
GOption[] opt0, opt1, opt2, opt3, opt4, opt5, opt6, opt7;
GToggleGroup opg0, opg1, opg2, opg3, opg4, opg5, opg6, opg7;
GDropList gdl; 
GTextField textIP;

PShader nebula;
PImage playImg, pauseImg, stopImg, prevImg, nextImg, visImg, pulseImg, saveImg;
PFont baseFont;

final int N_FBAND = 50, //кол-во частотных полос
          N_PWM_CH = 8, //кол-во каналов ШИМ
          SENS_BEAT = 150, //чувствительность бит-детектора, мс
          //кол-во каналов с одной частотной полосой (0..400, ~400..800, ~800..1200 Гц)
          N_FBAND_0 = 3,
          //параметры связи
          S_BAUD = 115200, //скорость подключения по COM-порту
          UDP_PORT = 8888, //порт UDP
          //характеристика регулировки, коэфф. корр. уровня
          KB_K0 = 20, //рег. громкости
          KB_K1 = 15, //рег. уровня каналов
          KB_K2 = 5,
          //значения смещений и размеров элементов
          X_OPT = 48, //смещение матрицы
          Y_OPT = 48,
          dX_OPT = 22, //смещение элемента в матрице
          dY_OPT = 20,
          X_KB = 250, //смещение группы регуляторов
          Y_KB = 43,
          dX_KB = 88, //смещение регулятора в группе
          dY_KB = 88,
          X_KB0 = 592, //смещение регулятора громкости
          Y_KB0 = 78,
          dY_KB_TXT = 12, //смещение подписи уровня
          KB_SIZE = 72, //размер регулятора уровня канала
          KB0_SIZE = 90, //размер регулятора громкости
          KB_TICKS = 10, //кол-во меток на регуляторах
          KB_RANG0 = 135, //положение меток
          KB_RANG1 = 45,
          
          BTN_SIZE = 69, //размер клавиш деки
          dY_BTN = 16, //смещение клавиш деки
          dX_BTN = 2,
          
          VIS_W = 720, //размеры окна визуализации эффекта Nebula, W x H
          VIS_H = 405;

//диапазоны частотных полос, объединяемых в каналы (5 каналов СЧ-ВЧ)
//~1,2..2 кГц, ~2..3,2 кГц, ~3,2..4.8 кГц, ~4,8..9,6 кГц, ~9,6..20 кГц
final int[] N_FBAND_L = {3, 5, 8, 12, 25},  //счет полос от 0
            N_FBAND_H = {4, 7, 11, 24, 49}; //N_FBAND_H[i] >= N_FBAND_L[i]
         
int idx = 0, //текущий индекс позиции файл-листа
    ch = 0, //текущий канал, 0 - аудио, 1...N_PWM_CH - каналы ШИМ
    ipn = 0, //текущий индекс массива COM-портов
    t, t_, t_val_, t_disp_;

String IP; //целевой IP-адрес для отправки данных UDP
//матрица соответствия частотных полос каналам 
String[][] fchArr = new String[N_PWM_CH][N_FBAND_0 + N_FBAND_L.length];  
//список соответствия частотных полос выбранному каналу
ArrayList<String> fchList = new ArrayList<String>(); 

byte out; //-128...127
byte[] udpData = new byte[12]; //пакет байт для отправки по UDP (без маркера конца кадра)

final float MIN_VALUE = 0, //мин/макс. знач. уровня по умолчанию = 0.75 (0...1.5) при логарифмической компресии или 0 при линейной
            MAX_VALUE = 5,  //> MIN_VALUE, 1.2 при логарифмической компресии или 5 (5...15) при линейной
            MIN_TIME = 100.0, //параметр time для шейдера Nebula
            MAX_TIME = 500.0, 
            DEF_TIME = 250.0, //default time
            MAX_BEAT_TIME = 3000.0, //макс. интервал бит-детектора, мс 
            //параметры регуляторов уровня
            KB_SENS = 0.8, //чувствительность
            KB_EASY = 3.0, //"скольжение"
            KB_GRIP = 0.89; //толщина линии
          
float[] a = new float[N_FBAND_0 + N_FBAND_L.length]; //массив уровней сигналов частотных полос
float[] ao = new float[N_PWM_CH]; //массив уровней сигналов каналов
float[] nebulaC = new float[N_PWM_CH]; //массив уровней сигналов для передачи в шейдер (для расширения, используются 1,2,6 элементы)
float[] g = new float[N_PWM_CH + 1]; //уровни коррекции по каналам (g[0] - канал звука)
float[] minVal = new float[N_PWM_CH]; //мин. знач. lg уровня по частотным каналам
float[] maxVal = new float[N_PWM_CH]; //макс. знач. lg уровня по частотным каналам

float nebulaTime,
      dT,
      dT0 = 1.4 * MAX_TIME,
      gain = 0.0; //уровень, -80...+14 dB

final String KB_GR_TITLE = "Channels Gain, dB",
             OPT_GR_TITLE = "Freq / Channel";
             //IP = "192.168.43.255";

String portName; //текущее имя COM-порта
       
String[] playList, //рабочий плейлист, очищенный от тегов
         playList0, //загруженный плейлист
         portNames, //массив имен системных COM-портов
         settings; //массив настроек (для расширения, используется запись IP)

boolean overPlayCtrl = false,
        overStopCtrl = false,
        overPrevCtrl = false,
        overNextCtrl = false,
        overVisCtrl = false,
        overSaveCtrl = false,
        trackChanged = false,
        manualPause = true,
        beatEn = true,
        ctrlEn = true,
        noList = false,
        noCon = false,
        noPort = false,
        visOn = false,
        saved = false;

void setup() {  
  size(720, 405); //создание канвы главного окна
  noStroke(); 
  //размещение контролов на канве
  //матрица (группы опций) соотв. частотных полос и каналов
  opg0 = new GToggleGroup();
  opt0 = new GOption[N_PWM_CH];
  for (int i = 0; i < N_PWM_CH; i++) {
    opt0[i] = new GOption(this, X_OPT+i*dX_OPT, Y_OPT+7*dY_OPT, 10, 10);
    opt0[i].tagNo = i;
    opg0.addControl(opt0[i]);
  }
  opt0[0].setSelected(true);
  
  opg1 = new GToggleGroup();
  opt1 = new GOption[N_PWM_CH];
  for (int i = 0; i < N_PWM_CH; i++) {
    opt1[i] = new GOption(this, X_OPT+i*dX_OPT, Y_OPT+6*dY_OPT, 10, 10);
    opt1[i].tagNo = 100+i;
    opg1.addControl(opt1[i]);
  }
  opt1[1].setSelected(true);
  
  opg2 = new GToggleGroup();
  opt2 = new GOption[N_PWM_CH];
  for (int i = 0; i < N_PWM_CH; i++) {
    opt2[i] = new GOption(this, X_OPT+i*dX_OPT, Y_OPT+5*dY_OPT, 10, 10);
    opt2[i].tagNo = 200+i;
    opg2.addControl(opt2[i]);
  }
  opt2[2].setSelected(true);
  
  opg3 = new GToggleGroup();
  opt3 = new GOption[N_PWM_CH];
  for (int i = 0; i < N_PWM_CH; i++) {
    opt3[i] = new GOption(this, X_OPT+i*dX_OPT, Y_OPT+4*dY_OPT, 10, 10);
    opt3[i].tagNo = 300+i;
    opg3.addControl(opt3[i]);
  }
  opt3[3].setSelected(true);
  
  opg4 = new GToggleGroup();
  opt4 = new GOption[N_PWM_CH];
  for (int i = 0; i < N_PWM_CH; i++) {
    opt4[i] = new GOption(this, X_OPT+i*dX_OPT, Y_OPT+3*dY_OPT, 10, 10);
    opt4[i].tagNo = 400+i;
    opg4.addControl(opt4[i]);
  }
  opt4[4].setSelected(true);
  
  opg5 = new GToggleGroup();
  opt5 = new GOption[N_PWM_CH];
  for (int i = 0; i < N_PWM_CH; i++) {
    opt5[i] = new GOption(this, X_OPT+i*dX_OPT, Y_OPT+2*dY_OPT, 10, 10);
    opt5[i].tagNo = 500+i;
    opg5.addControl(opt5[i]);
  }
  opt5[5].setSelected(true);
  
  opg6 = new GToggleGroup();
  opt6 = new GOption[N_PWM_CH];
  for (int i = 0; i < N_PWM_CH; i++) {
    opt6[i] = new GOption(this, X_OPT+i*dX_OPT, Y_OPT+dY_OPT, 10, 10);
    opt6[i].tagNo = 600+i;
    opg6.addControl(opt6[i]);
  }
  opt6[6].setSelected(true);
  
  opg7 = new GToggleGroup();
  opt7 = new GOption[N_PWM_CH];
  for (int i = 0; i < N_PWM_CH; i++) {
    opt7[i] = new GOption(this, X_OPT+i*dX_OPT, Y_OPT, 10, 10);
    opt7[i].tagNo = 700+i;
    opg7.addControl(opt7[i]);
  }
  opt7[7].setSelected(true);
  
  //регуляторы уровня с установкой опций
  //Цвета заливки (fill) и отрисовки (stroke) модифицированы в GKnob.java, с. 420-421, 429
  kb0 = new GKnob(this, X_KB0, Y_KB0, KB0_SIZE, KB0_SIZE, KB_SENS);
  kb0.setTurnMode(G4P.CTRL_ANGULAR);
  kb0.setEasing(KB_EASY);
  kb0.setNbrTicks(KB_TICKS);
  kb0.setGripAmount(KB_GRIP);
  kb0.setTurnRange(KB_RANG0, KB_RANG1);
  kb0.tagNo = 900;
  
  kb1 = new GKnob(this, X_KB, Y_KB, KB_SIZE, KB_SIZE, KB_SENS);
  kb1.setTurnMode(G4P.CTRL_ANGULAR);
  kb1.setEasing(KB_EASY);
  kb1.setNbrTicks(KB_TICKS);
  kb1.setGripAmount(KB_GRIP);
  kb1.setTurnRange(KB_RANG0, KB_RANG1);
  kb1.tagNo = 1000;
  
  kb2 = new GKnob(this, X_KB+dX_KB, Y_KB, KB_SIZE, KB_SIZE, KB_SENS);
  kb2.setTurnMode(G4P.CTRL_ANGULAR);
  kb2.setEasing(KB_EASY);
  kb2.setNbrTicks(KB_TICKS);
  kb2.setGripAmount(KB_GRIP);
  kb2.setTurnRange(KB_RANG0, KB_RANG1);
  kb2.tagNo = 2000;
  
  kb3 = new GKnob(this, X_KB+2*dX_KB, Y_KB, KB_SIZE, KB_SIZE, KB_SENS);
  kb3.setTurnMode(G4P.CTRL_ANGULAR);
  kb3.setEasing(KB_EASY);
  kb3.setNbrTicks(KB_TICKS);
  kb3.setGripAmount(KB_GRIP);
  kb3.setTurnRange(KB_RANG0, KB_RANG1);
  kb3.tagNo = 3000;
  
  kb4 = new GKnob(this, X_KB+3*dX_KB, Y_KB, KB_SIZE, KB_SIZE, KB_SENS);
  kb4.setTurnMode(G4P.CTRL_ANGULAR);
  kb4.setEasing(KB_EASY);
  kb4.setNbrTicks(KB_TICKS);
  kb4.setGripAmount(KB_GRIP);
  kb4.setTurnRange(KB_RANG0, KB_RANG1);
  kb4.tagNo = 4000;
  
  kb5 = new GKnob(this, X_KB, Y_KB+dY_KB, KB_SIZE, KB_SIZE, KB_SENS);
  kb5.setTurnMode(G4P.CTRL_ANGULAR);
  kb5.setEasing(KB_EASY);
  kb5.setNbrTicks(KB_TICKS);
  kb5.setGripAmount(KB_GRIP);
  kb5.setTurnRange(KB_RANG0, KB_RANG1);
  kb5.tagNo = 5000;
  
  kb6 = new GKnob(this, X_KB+dX_KB, Y_KB+dY_KB, KB_SIZE, KB_SIZE, KB_SENS);
  kb6.setTurnMode(G4P.CTRL_ANGULAR);
  kb6.setEasing(KB_EASY);
  kb6.setNbrTicks(KB_TICKS);
  kb6.setGripAmount(KB_GRIP);
  kb6.setTurnRange(KB_RANG0, KB_RANG1);
  kb6.tagNo = 6000;
  
  kb7 = new GKnob(this, X_KB+2*dX_KB, Y_KB+dY_KB, KB_SIZE, KB_SIZE, KB_SENS);
  kb7.setTurnMode(G4P.CTRL_ANGULAR);
  kb7.setEasing(KB_EASY);
  kb7.setNbrTicks(KB_TICKS);
  kb7.setGripAmount(KB_GRIP);
  kb7.setTurnRange(KB_RANG0, KB_RANG1);
  kb7.tagNo = 7000;
  
  kb8 = new GKnob(this, X_KB+3*dX_KB, Y_KB+dY_KB, KB_SIZE, KB_SIZE, KB_SENS);
  kb8.setTurnMode(G4P.CTRL_ANGULAR);
  kb8.setEasing(KB_EASY);
  kb8.setNbrTicks(KB_TICKS);
  kb8.setGripAmount(KB_GRIP);
  kb8.setTurnRange(KB_RANG0, KB_RANG1);
  kb8.tagNo = 8000;
  
  //установка шрифтов
  baseFont = createFont("Arial", 14);
  textFont(baseFont); //по умолчанию
  
  nebulaResetTime();

  //значения уровней по умолчанию 
  for(int i=0; i < N_PWM_CH+1; i++) g[i] = 0.0;
  for(int i=0; i < N_PWM_CH; i++) {
    minVal[i] = MIN_VALUE;
    maxVal[i] = MAX_VALUE;
    a[i] = 0.0;
    ao[i] = 0.0;
  }
  
  //заполнение матрицы соответствия по умолчанию
  for (int i=0; i < N_PWM_CH; i++)
    fchArr[i][0] = Integer.toString(i);

  //загрузка файлов ресурсов
  //(находятся в каталоге /data скетча)
  try {
    playImg = loadImage("play.png");
    pauseImg = loadImage("pause.png");
    stopImg = loadImage("stop.png");
    nextImg = loadImage("next.png");
    prevImg = loadImage("prev.png");
    visImg = loadImage("visual.png");
    pulseImg = loadImage("pulse.png");
    saveImg = loadImage("save_.png");
    } catch (RuntimeException e) {
  }
  
  minim = new Minim(this);
  
  //файл плейлиста
  try {
    playList0 = loadStrings("playlist.m3u8");
    //printArray(playList0);
    } catch (RuntimeException e) {
  }
  
  if (playList0 != null) { //файл плейлиста загружен 
      //удаление строк комментариев из массива
      //создание "чистого" списка
      playList = new String[0];
      for (int i=0; i < playList0.length; i++) {
        if (!playList0[i].contains("#EXT")) playList = append(playList, playList0[i]); 
      }
      //printArray(playList);
      //println(playList.length);      
      player = minim.loadFile(playList[idx], 1024);
      fft = new FFT(player.bufferSize(), player.sampleRate());  
      //линейное разделение частотных полос
      fft.linAverages(N_FBAND);
      //бит-детектор / пик-детектор
      beat = new BeatDetect(player.bufferSize(), player.sampleRate());
      //beat = new BeatDetect(); //пик-детектор
      beat.setSensitivity(SENS_BEAT);
      bl = new BeatListener(beat, player); //отключить для пик-детектора
      //чтение мета-данных
      meta = player.getMetaData();
      //плейлист загружен
      println("Playlist is loaded");
  } else {
      //ошибка загрузки плейлиста
      noList = true;
      println("Playlist is NOT loaded");
  }
  
  //Список COM-портов   
  try {
    printArray(Serial.list());
    portNames = Serial.list();
    portName = portNames[0];
    } catch (ArrayIndexOutOfBoundsException e) {
    noPort = true;
    portNames = append(portNames, "No COM-port");
    println("No COM-port");
  }
  
  //селектор порта
  gdl = new GDropList(this, X_OPT, Y_KB+2*dY_KB+dY_BTN, 143, 130, 5);
  gdl.setLocalColorScheme(6);
  gdl.setItems(portNames, 0); //поз.0 по умолчанию
  if (!noPort) gdl.addEventHandler(this, "portSelect"); 
  //связь с исполнительным устройством 
  //через USB COM-порт по умолчанию
  try {
      portName = gdl.getSelectedText();
      port = new Serial(this, portName, S_BAUD);
      noCon = false;
      println("Port " + portName + " is opened");
      } catch (RuntimeException e) {
      noCon = true;
      println("No device connected"); 
    }
  //поле ввода IP  
  textIP = new GTextField(this, X_OPT, Y_KB+2*dY_KB+dY_BTN+BTN_SIZE-32, 143, 21);
  //файл настроек
  try {
    settings = loadStrings("settings.cfg");
    //printArray(settings);
    IP = settings[0];
    } catch (RuntimeException e) {
  }
  textIP.setText(IP);
    
  //широковещательное подключение на порту UDP_PORT
  udp = new UDP(this, UDP_PORT);
  
  //по умолчанию, если IP-адрес не указан, UDP 
  //отправляет пакеты на подключенные адреса группы на текущем порту
  //udp.send(data); // = send( data, group_ip, port ); //the data type is byte
  //при создании мультикаст-подключения сохраняется возможность отправки данных
  //на определенный IP-адрес (напр., send( "the messsage", "192.168.0.2", 7010);

  println("init broadcast socket... " + udp.isBroadcast());
  udp.log(true);
}

void draw() {
  background(0);
  t = millis();

  //Handler отрисовки окна визуализации
  if (visOn) window.addDrawHandler(this, "windowDraw");
  
  if (!noList) { //если загружен плейлист, загрузить все функции плеера
    //установка уровня громкости при смене трека
    if (trackChanged) {
      player.setGain(g[0]);
      trackChanged = false;
    }
    
    //получение уровней сигналов частотных полос
    fft.forward(player.mix);
    //объединение в диапазоны
    if (player.isPlaying()) {
      //первые N_FBAND_0 каналов с одной частотной полосой
      for (int i = 0; i < N_FBAND_0; i++) a[i] = fft.getAvg(i);
      //каналы N_FBAND_L..N_FBAND_5
      for (int j = 0; j < N_PWM_CH-N_FBAND_0; j++) {//ok!
        a[j+N_FBAND_0] = 0.0; //обнуление уровня полосы
        for (int i = N_FBAND_L[j]; i < N_FBAND_H[j]+1; i++) //ok!
          if (fft.getAvg(i) > a[j+N_FBAND_0]) a[j+N_FBAND_0] = fft.getAvg(i); //амплитуды частотных полос, MAX value
        }
      } else //player is stopped or paused
        for (int i = 0; i < N_PWM_CH; i++) a[i] = 0.0; //обнуление уровней полос    
      //амплитуды полос вычислены
      //назначение полос каналам согласно матрице      
      for (int i=0; i < N_PWM_CH; i++) { //перебор каналов
        ao[i] = 0.0; //обнуление уровня канала
        for (int j=0; j < N_FBAND_0 + N_FBAND_L.length; j++) //перебор полос
          if (Arrays.asList(fchArr[i]).contains(Integer.toString(j))) //найдено соотв. номера полосы в массиве канала
            if (a[j] > ao[i]) ao[i] = a[j]; //уровень сигнала полосы выше текущего уровня сигнала канала, MAX value
      }
      
      for (int i = 0; i < N_PWM_CH; i++) {
        //println(g[i+1]); //вывод значений корр. уровней
        //println(a[i]); //вывод значений уровней каналов
        //логарифмическое сжатие уровней каналов
        //log(avg*10^(g/20)) = log(avg) + g/20, g[0] - уровень аудио
        //if (ao[i] > 0) ao[i] = log(ao[i]) + (g[i+1]/20);
          //линейная компрессия
          ao[i] = ao[i]*pow(10, g[i+1]/20);
          
          //динамическая коррекция диапазона
          if (ao[i] < minVal[i]) minVal[i] = ao[i];
            else if (ao[i] > maxVal[i]) maxVal[i] = ao[i];
            
          ao[i] = constrain(ao[i], minVal[i], maxVal[i]);
          ao[i] = norm(ao[i], minVal[i], maxVal[i]);
          //нормализация в интервал (-127...127) для побайтовой отправки в USB COM-порт
          out = (byte)round(map(ao[i], 0.0, 1.0, -127.0, 127.0));
          if (!noCon) port.write(out); //отправка данных
          udpData[i] = out;
          print(out+":");
          //вкл. подсветки, если стоп / пауза
          if (!player.isPlaying() || manualPause) ao[i] = -127;
      }
      
    //бит-детектор с определением типа инструментов
    //beat.detect(player.mix); //режим пик-детектора
    
    if (beat.isKick()) { //isOnset() //пик-детектор
      dT = t - t_;
      dT = constrain(dT, float(SENS_BEAT), MAX_BEAT_TIME);
      t_ = t;
      out = (byte)127; //Kick detected
    } else out = (byte)0; //no Kick detected
    if (!noCon) port.write(out);
    udpData[N_PWM_CH] = out;
    print(out+":");
    
    if (beat.isSnare()) out = (byte)127; //Snare detected
      else out = (byte)0; //no Snare detected
    if (!noCon) port.write(out);
    udpData[N_PWM_CH+1] = out;
    print(out+":");
    
    if (beat.isHat()) out = (byte)127; //Hat detected
      else out = (byte)0; //no Hat detected
    if (!noCon) port.write(out);
    udpData[N_PWM_CH+2] = out;
    print(out+":");
    
    //println(dT);
    out = (byte)round(map(dT, float(SENS_BEAT), MAX_BEAT_TIME, -127.0, 127.0));
    if (!noCon) port.write(out);
    udpData[N_PWM_CH+3] = out;
    print(out+":");
    
    out = (byte)-128; //маркер конца кадра
    if (!noCon) port.write(out);
    println(out);
    
    //UDP broadcast
    udp.send(udpData, IP, UDP_PORT);
    
    //сброс minVal, maxVal по dT 
    if (t > t_val_) {
      t_val_ =  t + round(dT);
      for(int i=0; i < N_PWM_CH; i++) {
        minVal[i] = MIN_VALUE;
        maxVal[i] = MAX_VALUE;
      }
      //println("reset comp. gates");
    }
    
    //плавное изменение и ограничение ритм-параметра для шейдера
    if (dT > dT0) dT0 = 1.001*dT0;
    if (dT < dT0) dT0 = 0.999*dT0;
    dT0 = constrain(dT0, float(SENS_BEAT), MAX_BEAT_TIME);
    
    float kT = map(dT0, float(SENS_BEAT), MAX_BEAT_TIME, MIN_TIME, MAX_TIME);  
    //println(kT);
    
    if (player.isPlaying() && beatEn) nebulaTime = millis() / kT;
      else nebulaResetTime();
    
    if (player.isPlaying() && ctrlEn) {
      nebulaC[1] = ao[1];
      nebulaC[2] = ao[2];
      nebulaC[3] = ao[6];
    }
  
  //вывод деки управления и текстовой информации
  if (player.isPlaying()) image(pauseImg, X_KB+4*dX_KB+dX_BTN, Y_KB+2*dY_KB+dY_BTN, BTN_SIZE, BTN_SIZE);
    else image(playImg, X_KB+4*dX_KB+dX_BTN, Y_KB+2*dY_KB+dY_BTN, BTN_SIZE, BTN_SIZE);
  image(stopImg, X_KB+3*dX_KB+dX_BTN, Y_KB+2*dY_KB+dY_BTN, BTN_SIZE, BTN_SIZE);
  image(prevImg, X_KB+dX_KB+dX_BTN, Y_KB+2*dY_KB+dY_BTN, BTN_SIZE, BTN_SIZE);
  image(nextImg, X_KB+2*dX_KB+dX_BTN, Y_KB+2*dY_KB+dY_BTN, BTN_SIZE, BTN_SIZE);
  image(visImg, X_KB+dX_BTN, Y_KB+2*dY_KB+dY_BTN, BTN_SIZE, BTN_SIZE);
  image(pulseImg, X_KB+4*dX_KB+dX_BTN, Y_KB+2*dY_KB+BTN_SIZE+3*dY_BTN, BTN_SIZE, 35);
  image(saveImg, X_OPT+150, Y_KB+2*dY_KB+dY_BTN+BTN_SIZE-32, 21, 21);

  textFont(baseFont);

  text(meta.fileName(), X_OPT, Y_KB+2*dY_KB+BTN_SIZE+3*dY_BTN);
  text(meta.title(), X_OPT, Y_KB+2*dY_KB+BTN_SIZE+4*dY_BTN);
  
  int totalMinutes = (int)(meta.length()/(1000*60)) % 60;
  int totalSeconds = (int)(meta.length()/1000) % 60;
  int playheadMinutes = (int)(player.position()/(1000*60)) % 60;
  int playheadSeconds = (int)(player.position()/1000) % 60;
  text(playheadMinutes + ":" + nf(playheadSeconds, 2) + " / " + totalMinutes + ":" + nf(totalSeconds, 2), X_OPT, Y_KB+2*dY_KB+BTN_SIZE+5*dY_BTN);
  
  //переход к след. треку по окончании текущего
  //if (player.position() == player.length()) { //стандартный метод не всегда работает правильно!
  if (!player.isPlaying() && !manualPause) {
    player.close();
        if (idx < playList.length - 1) idx++; else idx = 0; //в начало плейлиста       
        player = minim.loadFile(playList[idx], 1024);
        fft = new FFT(player.bufferSize(), player.sampleRate());
        fft.linAverages(N_FBAND);
        beat = new BeatDetect(player.bufferSize(), player.sampleRate());
        beat.setSensitivity(SENS_BEAT);
        bl = new BeatListener(beat, player);
        meta = player.getMetaData();
        nebulaResetTime();
        player.play();
        //t_meta_ = t + T_DISP_META;
        trackChanged = true;
  }
  
  //определение позиции курсора
  if (mouseX > X_KB+4*dX_KB+dX_BTN && mouseX < X_KB+4*dX_KB+dX_BTN+BTN_SIZE 
    && mouseY > Y_KB+2*dY_KB+dY_BTN && mouseY < Y_KB+2*dY_KB+dY_BTN+BTN_SIZE) 
      overPlayCtrl = true; else overPlayCtrl = false;
      
  if (mouseX > X_KB+3*dX_KB+dX_BTN && mouseX < X_KB+3*dX_KB+dX_BTN+BTN_SIZE
    && mouseY > Y_KB+2*dY_KB+dY_BTN && mouseY < Y_KB+2*dY_KB+dY_BTN+BTN_SIZE) 
      overStopCtrl = true; else overStopCtrl = false;

  if (mouseX > X_KB+dX_KB+dX_BTN && mouseX < X_KB+dX_KB+dX_BTN+BTN_SIZE
    && mouseY > Y_KB+2*dY_KB+dY_BTN && mouseY < Y_KB+2*dY_KB+dY_BTN+BTN_SIZE)
      overPrevCtrl = true; else overPrevCtrl = false;
 
  if (mouseX > X_KB+2*dX_KB+dX_BTN && mouseX < X_KB+2*dX_KB+dX_BTN+BTN_SIZE
    && mouseY > Y_KB+2*dY_KB+dY_BTN && mouseY < Y_KB+2*dY_KB+dY_BTN+BTN_SIZE)
      overNextCtrl = true; else overNextCtrl = false;
 
  if (mouseX > X_KB+dX_BTN && mouseX < X_KB+dX_BTN+BTN_SIZE 
    && mouseY > Y_KB+2*dY_KB+dY_BTN && mouseY < Y_KB+2*dY_KB+dY_BTN+BTN_SIZE)
      overVisCtrl = true; else overVisCtrl = false;
      
  } else { //noList, плейлист не загружен
    nebulaResetTime();
    
    text("Playlist is NOT loaded", X_OPT, Y_KB+2*dY_KB+BTN_SIZE+3*dY_BTN);
  }
  
  //кнопка Сохранить IP
  if (mouseX > X_OPT+150 && mouseX < X_OPT+171 
    && mouseY > Y_KB+2*dY_KB+dY_BTN+BTN_SIZE-32 && mouseY < Y_KB+2*dY_KB+dY_BTN+BTN_SIZE-11)
      overSaveCtrl = true; else overSaveCtrl = false;
      
  //заголовки секций управления
  text(OPT_GR_TITLE, X_OPT+35, Y_OPT-13);
  text(KB_GR_TITLE, X_KB+109, Y_KB-9);
  text('F', X_OPT-16, Y_OPT-13);
  text("Ch", X_OPT+167, Y_KB+dY_KB+KB_SIZE+dY_KB_TXT);
  //нумерация секции выбора частот-каналов
  for(int i=0; i < N_PWM_CH; i++) //частоты
      text(N_FBAND_0 + N_FBAND_L.length-i, X_OPT-16, Y_OPT+10+20*i);
  for(int i=0; i < N_PWM_CH; i++) //каналы
      text(i+1, X_OPT+22*i, Y_KB+dY_KB+KB_SIZE+dY_KB_TXT);
  
  //вывод значений коррекции уровня по каналам
  text(String.format("%.1f", g[0]), X_KB+4*dX_KB, Y_KB+dY_KB+KB_SIZE+dY_KB_TXT); //audio
  
  for(int i=0; i < N_PWM_CH/2; i++) 
      text(String.format("%.1f", g[i+1]), X_KB+i*dX_KB, Y_KB+KB_SIZE+dY_KB_TXT);
  for(int i=N_PWM_CH/2; i < N_PWM_CH; i++) 
      text(String.format("%.1f", g[i+1]), X_KB+(i-N_PWM_CH/2)*dX_KB, Y_KB+dY_KB+KB_SIZE+dY_KB_TXT);
  //сообщение о сохранении настроек
  if (saved) {
    fill(0, 255, 0);
    text("Saved", X_OPT, Y_KB+2*dY_KB+dY_BTN+BTN_SIZE+2);
    if (t > t_disp_) saved = false;
  }
  
  //индикатор соcтояния порта
  if (!noCon) fill(0, 255, 0);
    else fill(255, 0, 0);
  rect(X_OPT+155, Y_KB+2*dY_KB+dY_BTN+5, 10, 10);
  
  fill(255);  
}

//обработка нажатий на клавишу "мыши"
void mousePressed() {
  if (!noList) { //по позиции курсора   
    if (overPlayCtrl) { 
       if (player.isPlaying()) {
          player.pause();
          manualPause = true;
        } else {
          player.play();
          player.setGain(g[0]);
          manualPause = false;
        }
    }    
    if (overNextCtrl) {
          player.close();
          if (idx < playList.length - 1) idx++; else idx = 0; //в начало плейлиста        
          player = minim.loadFile(playList[idx], 1024);
          fft = new FFT(player.bufferSize(), player.sampleRate());
          fft.linAverages(N_FBAND);
          beat = new BeatDetect(player.bufferSize(), player.sampleRate());
          beat.setSensitivity(SENS_BEAT);
          bl = new BeatListener(beat, player);
          meta = player.getMetaData();
          nebulaResetTime();
          if (!manualPause) player.play();
          trackChanged = true;
    }    
    if (overPrevCtrl) {
          player.close();
          if (idx > 0) idx--; else idx = 0; //в начало плейлиста       
          player = minim.loadFile(playList[idx], 1024);
          fft = new FFT(player.bufferSize(), player.sampleRate());
          fft.linAverages(N_FBAND);
          beat = new BeatDetect(player.bufferSize(), player.sampleRate());
          beat.setSensitivity(SENS_BEAT);
          bl = new BeatListener(beat, player);
          meta = player.getMetaData();
          nebulaResetTime();
          if (!manualPause) player.play();
          trackChanged = true;
    }    
    if (overStopCtrl) {
          player.close();
          idx = 0; //в начало плейлиста        
          player = minim.loadFile(playList[idx], 1024);
          fft = new FFT(player.bufferSize(), player.sampleRate());
          fft.linAverages(N_FBAND);
          beat = new BeatDetect(player.bufferSize(), player.sampleRate());
          beat.setSensitivity(SENS_BEAT);
          bl = new BeatListener(beat, player);
          meta = player.getMetaData();
          nebulaResetTime();
          manualPause = true;
    }
    if (overVisCtrl && !visOn) createWindows(); //создание окна для визуализации шейдера 
    if (overSaveCtrl) {
      IP = textIP.getText(); //назначение IP для отправки данных по UDP
      settings[0] = IP;
      saveStrings("data/settings.cfg", settings); //Сохранить IP в файле настроек
      //G4P.showMessage(this, "Settings are saved", "", G4P.INFO);
      saved = true;
      t_disp_ = t + 2000;
    }
  } //end noList
}

//обработка событий клавиатуры
void keyPressed() {
  if (key == 'b') beatEn = !beatEn; //откл. передачи ритм-данных в шейдер
  if (key == 'c') ctrlEn = !ctrlEn; //откл. передачи амплитуд в шейдер
  if (key == 'v' && !visOn) createWindows(); //создание окна для визуализации шейдера
}

//получение значений коррекции от регуляторов
//общий handler по умолчанию для всех регуляторов 
//в один момент может изм. полож. только одного регулятора (допущение / упрощение кода)
public void handleKnobEvents(GValueControl knob, GEvent event) {
  int kn = knob.tagNo / 1000;
  
  switch(kn) {
    case 0: //audio
      g[0] = KB_K0*log(2*knob.getValueF());
      if (player.isPlaying()) player.setGain(g[0]);
      break;
    default:
      if (knob.getValueF() >= 0) 
        g[kn] = KB_K1*log(2*knob.getValueF());
        else g[kn] = KB_K2*log(2*knob.getValueF());
    }
}

//обработчик событий селектора опций
//общий handler по умолчанию для всех групп, аналогично регуляторам
public void handleToggleControlEvents(GToggleControl option, GEvent event) {
  int chNum = option.tagNo % 100;
  String fbNum = Integer.toString(floor(option.tagNo / 100));
  println(chNum+":"+fbNum);
  
  //каждому каналу может быть назначено несколько полос,
  //но каждая частотная полоса может быть назначена только одному каналу
  //поэтому после каждого изм. ищем в матрице
  //ранее назначенное соотв. данной полосы и удаляем его
  //после чего назначаем новое и пересохраняем матрицу
  
  //хранение матрицы в двумерном массиве для удобства работы с индексами
  //построчное преобразование во временный список для удобства обработки элементов
  //(удаление, добавление)
  
  //чтение матрицы соотв. полос и каналов в список
  //фильтр по null
  //перебор каналов на соотв. искомой полосе
  for (int i=0; i < N_PWM_CH; i++) {
    fchList.clear();
    for (int j=0; j < N_FBAND_0 + N_FBAND_L.length; j++) {
      if (fchArr[i][j] != null)
        fchList.add(fchArr[i][j]);
    }
    println(fchList);
    if (fchList.contains(fbNum)) //соотв. найдено
      fchList.remove(fbNum);      
    //println("list:"+fchList);
    
    //очистка массива
    for (int j=0; j < fchArr[i].length; j++)
        fchArr[i][j] = null;
    //обратное преобразование списка в массив 
    for (int j=0; j < fchList.size(); j++)
      fchArr[i][j] = fchList.get(j);
    //print("array:"); println(fchArr[i]);
  }
  //назначение соотв. выбранных канала и полосы
  fchList.clear();
  for (int j=0; j < N_FBAND_0 + N_FBAND_L.length; j++)
      if (fchArr[chNum][j] != null)
        fchList.add(fchArr[chNum][j]);

  //добавление номера полосы (с доп. проверкой на наличие в списке)
  if (!fchList.contains(fbNum)) fchList.add(fbNum);
  //println("new list:"+fchList);
  //очистка массива
  for (int j=0; j < fchArr[chNum].length; j++)
      fchArr[chNum][j] = null;
  //обратное преобразование в массив
  for (int j=0; j < fchList.size(); j++)
      fchArr[chNum][j] = fchList.get(j);  

  //вывод массива
  //println(fchArr[chNum]);
  //вывод матрицы
  //for (int j=0; j < N_PWM_CH; j++) println(fchArr[j]);
}

//обработчик событий селектора портов
public void portSelect(GDropList source, GEvent event) {
  portName = source.getSelectedText();
      //остановить текущее соединение
      if (!noCon) {
        port.clear();
        port.stop();
        println("COM-port is closed");
      }
      //открыть новый USB COM-порт
      try {
        port = new Serial(this, portName, 115200);
        noCon = false;
        println("Port " + portName + " is opened");
      } catch (RuntimeException e) {
        noCon = true;
        println("No device connected");
      }
}

//создание окна визуализации
public void createWindows() {
    window = GWindow.getWindow(this, "Nebula", 30, 30, VIS_W, VIS_H, P2D);
    window.setActionOnClose(window.CLOSE_WINDOW);
    window.addDrawHandler(this, "windowSetup");
    
    visOn = true;
    delay(500); //задержка на загрузку handler'а окна, 50-1000 ms
}

//Handler создания окна визуализации
public void windowSetup(PApplet appc, GWinData data) {
  appc.background(0);  
  nebula = appc.loadShader("nebula.glsl");
  nebula.set("resolution", window.width, window.height);
  //заполнение массива данных для передачи шейдеру случайными значениями
  nebulaC[1] = random(1.0);
  nebulaC[2] = random(1.0);
  nebulaC[3] = random(1.0);
}

//Handler отрисовки окна визуализации
public void windowDraw(PApplet appc, GWinData data) {
  appc.background(0);
  nebula.set("time", nebulaTime); 
  nebula.set("c1", nebulaC[1]);
  nebula.set("c2", nebulaC[2]);
  nebula.set("c3", nebulaC[3]);
  appc.shader(nebula);  
  appc.rect(0, 0, window.width, window.height);
}

//класс обработчика бит-детектора
class BeatListener implements AudioListener {
  private BeatDetect beat;
  private AudioPlayer source;
  
  BeatListener(BeatDetect beat, AudioPlayer source) {
    this.source = source;
    this.source.addListener(this);
    this.beat = beat;
  }  
  void samples(float[] samps) {
    beat.detect(source.mix);
  }  
  void samples(float[] sampsL, float[] sampsR) {
    beat.detect(source.mix);
  }
}

void nebulaResetTime() {
  nebulaTime = millis() / DEF_TIME;
  dT0 = 1.4 * MAX_TIME;
}

//закрытие экземпляров плеера и minim при закрытии приложения
void stop() {
 if (player != null) player.close();
 minim.stop(); 
 super.stop();
}
