/*
* Программа удаленного контроллера
* Arduino Mega 2560
* для управления ЦМУ Nebulizer Pro
* Разработка: Панкратьев Д. И., 2020
*/
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
/*
 * Ссылка для мнеджера плат:
 * http://arduino.esp8266.com/stable/package_esp8266com_index.json
 */
#ifndef STASSID
#define STASSID "ID"
#define STAPSK  "PASSWORD"
#endif

unsigned int localPort = 8888; //локальный порт прослушивания UDP

char packetBuffer[UDP_TX_PACKET_MAX_SIZE + 1]; //буфер хранения полученных данных

WiFiUDP Udp;

/** Ch - DX - GPIO
 *  0 PWM - TX - 1
 *  1 PWM - RX - 3
 *  2 PWM - D8 - 15
 *  3 PWM - D7 - 13
 *  4 PWM - D6 - 12
 *  5 PWM - D5 - 14
 *  6 PWM - D4 - 2
 *  7 PWM - D3 - 0
 *  8 PWM - D2 - 4 (ритм-канал)
 *  1,2 BIT PWM - D1 - 5 (BIT1_PWM_OUT при сигнале бит-канала 1, 255 при сигнале бит-канала 2, BCL_PWM_OUT при подсветке)
 *  0 BIT - D0 - 16
 */

const uint8_t AI_PIN[] = {1, 3, 15, 13, 12, 14, 2, 0}, //массив номеров GPIO ШИМ частотных полос
          BEAT0_PIN = 16, //выход бит-детектора
          BEAT12_PIN = 5,         
          RTM_PIN = 4, //выход ШИМ ритм-канала
          IN_ARR_SIZE = 12, //размер входного массива данных
          MIN_PWM_LED = 0, //мин. коэфф. ШИМ (<255); 0-16 - для LED , 16-64 - для ламп накал.
          MIN_PWM_OUT = 50, //мин. коэфф. ШИМ ритм-канала
          BIT1_PWM_OUT = 128, //255 max
          BCL_PWM_OUT = 81; //уровень подсветки, 255 max

const uint16_t CON_WAIT = 180, //интервал ожидания подключения Wi-Fi (до перехода в авторежим эмуляции), с (от 2000 при настройке)
          SIM_INT1 = 100, //мин/макс. интервал симуляции, мс
          SIM_INT2 = 2000,
          SIM_BIT = 50, //длительность сигнала бит-канала при симуляции, мс
          F_PWM = 100; //частота ШИМ

int8_t in[IN_ARR_SIZE]; //входной массив данных, char(-128...127) 
/** входные значения сигналов:
 * in[0]...in[7] - частотных полос
 * in[8]...in[10] - 0-2 бит-каналов (-127/127)
 * in[11] - ритм-канала
 */

uint16_t c; //счетчик циклов ожидания
bool noCon = false;
unsigned long t, t_, t__;

void setup() {  
  analogWriteRange(255); //1023 default
  analogWriteFreq(F_PWM); //1 kHz default
  
  pinMode(BEAT0_PIN, OUTPUT);
  
  //уст. знач. -127 по умолч. для элем. вх. массива
  for(uint8_t i=0; i<IN_ARR_SIZE; i++) in[i] = -127;
  //начальный уровень PWM
  for(uint8_t i=0; i<IN_ARR_SIZE-4; i++) analogWrite(AI_PIN[i], 0);
  //начальный уровень ритм-канала
  analogWrite(RTM_PIN, 0);
  analogWrite(BEAT12_PIN, 0);
  //настройка макс. тока ШИМ
  /*
  for (uint8_t i=0; i < IN_ARR_SIZE - 4; i++){
    analogWrite(AI_PIN[i], 255);
  }
  //настройка макс. тока бит-каналов
  digitalWrite(BEAT0_PIN, HIGH);
  digitalWrite(BEAT1_PIN, HIGH);
  digitalWrite(BEAT2_PIN, HIGH);
  */
  //настройка мин. тока ШИМ
  /*
  for (uint8_t i=0; i < IN_ARR_SIZE - 4; i++){
    analogWrite(AI_PIN[i], MIN_PWM_LED);
  }
  //настройка мин. тока ритм-канала 
  analogWrite(RTM_PIN, MIN_PWM_OUT);
  */
  /*
  Serial.begin(115200);
  Serial.println("\nStarting...");
  */
  WiFi.mode(WIFI_STA);
  WiFi.begin(STASSID, STAPSK);
  //ожидание подключения
  while (WiFi.status() != WL_CONNECTED) { //мигание бит-канала 0
    digitalWrite(BEAT0_PIN, !digitalRead(BEAT0_PIN));
    //digitalWrite(BEAT0_PIN, HIGH); //при настройке
    delay(250); 
    c++;
    //переход в авторежим эмуляции по истечении интервала CON_WAIT
    if (c == CON_WAIT * 4) { 
      noCon = true;
      break; 
    }
  }
  digitalWrite(BEAT0_PIN, LOW);
  //подсветка с уровнем BCL_PWM_OUT на выходе BEAT12_PIN
  analogWrite(BEAT12_PIN, BCL_PWM_OUT);
  //старт сессии UDP
  if (!noCon) Udp.begin(localPort);
}

void loop() {
  t = millis();
  
  if (!noCon) { //при наличии связи
    //чтение данных при их доступности
      int packetSize = Udp.parsePacket();
      if (packetSize) {
        //read the packet into packetBufffer
        int n = Udp.read(packetBuffer, UDP_TX_PACKET_MAX_SIZE);
        //packetBuffer[n] = 0x00;
        //заполнение входного массива данными
        for (uint8_t i=0; i<n; i++) in[i] = (signed char)packetBuffer[i];
      } 
   } else { //в отсутствие связи
      if (t_ < t) {
        t_ = t + random(SIM_INT1, SIM_INT2);
        //заполнение входного массива случайными числами
        for (uint8_t i=0; i<IN_ARR_SIZE-4; i++) in[i] = (signed char)random(-127, 127); //каналы ШИМ
        if (random(-127, 127)%2 == 0) in[IN_ARR_SIZE-4] = 0; else in[IN_ARR_SIZE-4] = 127; //бит-каналы
        if (random(-127, 127)%2 == 0) in[IN_ARR_SIZE-3] = 0; else in[IN_ARR_SIZE-3] = 127;
        if (random(-127, 127)%2 == 0) in[IN_ARR_SIZE-2] = 0; else in[IN_ARR_SIZE-2] = 127;
        in[IN_ARR_SIZE-1] = (signed char)random(-127, 127); //ритм-канал
        }
       if (t__ < t) { //сброс сигнала бит-каналов
        t__ = t + SIM_BIT;
        in[IN_ARR_SIZE-4] = 0;
        in[IN_ARR_SIZE-3] = 0;
        in[IN_ARR_SIZE-2] = 0;
       }
      }
        //вых. сигнал ШИМ частотных полос,
        //map в диапазон (MIN_PWM_LED... 255)
        for (uint8_t i=0; i<IN_ARR_SIZE-4; i++) analogWrite(AI_PIN[i], map(in[i], -127, 127, MIN_PWM_LED, 255));
        //бит-детектор, канал 0
        if (in[IN_ARR_SIZE-4] == 127) digitalWrite(BEAT0_PIN, HIGH); else digitalWrite(BEAT0_PIN, LOW);
        //бит-детектор, каналы 1-2
        if (in[IN_ARR_SIZE-3] == 127) analogWrite(BEAT12_PIN, BIT1_PWM_OUT);
          else if (in[IN_ARR_SIZE-2] == 127) analogWrite(BEAT12_PIN, 255);
            else analogWrite(BEAT12_PIN, 0);
        //вых сигнал ШИМ ритм-канала, 
        //map в "обратный" диапазон (255... MIN_PWM_OUT)
        analogWrite(RTM_PIN, map(in[IN_ARR_SIZE-1], -127, 127, 255, MIN_PWM_OUT));
        //подсветка с уровнем BCL_PWM_OUT на выходе BEAT12_PIN
        if (in[0] == -127 && in[1] == -127 && in[2] == -127 && in[3] == -127 &&
          in[4] == -127 && in[5] == -127 && in[6] == -127 && in[7] == -127) {
            analogWrite(RTM_PIN, 0);
            analogWrite(BEAT12_PIN, BCL_PWM_OUT);
        }
}
