/*
* Программа локального контроллера
* Arduino Mega 2560
* для управления ЦМУ Nebulizer Pro
* Разработка: Панкратьев Д. И., 2020
*/

//8 каналов ШИМ, 3 канала бит-детектора (Kick, Snare, Hat), ритм-канал ШИМ

const uint8_t AI_PIN[] = {4, 5, 6, 7, 8, 9, 10, 11}, //массив номеров выходов ШИМ частотных полос
          BEAT0_PIN = 22, //выходы бит-детектора
          BEAT1_PIN = 23,
          BEAT2_PIN = 24,          
          RTM_PIN = 12, //выход ШИМ ритм-канала
          BCL_PIN_IDX = 6, //индекс ШИМ-выхода подсветки в AI_PIN[]
          IN_ARR_SIZE = 12, //размер входного массива данных
          MIN_PWM_LED = 0, //мин. коэфф. ШИМ (<255); 0-16 - для LED , 16-64 - для ламп накал.
          MIN_PWM_OUT = 50, //мин. коэфф. ШИМ ритм-канала
          BCL_PWM_OUT = 81; //уровень подсветки, 255 max

const uint16_t SIM_WAIT = 180, //интервал ожидания до перехода в авторежим эмуляции, с
          SIM_INT1 = 100, //мин/макс. интервал симуляции, мс
          SIM_INT2 = 2000,
          SIM_BIT = 50; //длительность сигнала бит-канала при симуляции, мс

char in[IN_ARR_SIZE]; //входной массив данных, char(-128...127) 
/* входные значения сигналов:
 * in[0]...in[7] - частотных полос
 * in[8]...in[10] - бит-каналов (-127/127)
 * in[11] - ритм-канала
 */

uint8_t n = 0;
unsigned long t, t_s_, t_, t__;
boolean rComplete = false,
        noCon = false;

void setup() {
  Serial.begin(115200);
  pinMode(BEAT0_PIN, OUTPUT);
  pinMode(BEAT1_PIN, OUTPUT);
  pinMode(BEAT2_PIN, OUTPUT);
  digitalWrite(BEAT0_PIN, LOW);
  digitalWrite(BEAT1_PIN, LOW);
  digitalWrite(BEAT2_PIN, LOW);
  //уст. знач. -127 по умолч. для элем. вх. массива
  for(uint8_t i=0; i<IN_ARR_SIZE; i++){
    in[i] = -127;
    }
  //начальный уровень ритм-канала
  analogWrite(RTM_PIN, 0);
  //включение подсветки
  analogWrite(AI_PIN[BCL_PIN_IDX], BCL_PWM_OUT);
  //настройка макс. тока ШИМ
  /*
  for (uint8_t i=0; i < IN_ARR_SIZE - 4; i++){
    analogWrite(AI_PIN[i], 255);
  }
  //настройка макс. тока бит-каналов
  digitalWrite(BEAT0_PIN, HIGH);
  digitalWrite(BEAT1_PIN, HIGH);
  digitalWrite(BEAT2_PIN, HIGH);
  */
  //настройка мин. тока ШИМ
  /*
  for (uint8_t i=0; i < IN_ARR_SIZE - 4; i++){
    analogWrite(AI_PIN[i], MIN_PWM_LED);
  }
  //настройка мин. тока ритм-канала 
  analogWrite(RTM_PIN, MIN_PWM_OUT);
  */
}

void loop() {            
  //закомментировать тело главного цикла при настройке
  t = millis();
  if (Serial.available() > 0 && !rComplete) {
      //получение последовательности вх. данных
      char inChar = Serial.read();
      if(inChar == -128){ //маркер кадра
        rComplete = true;
        n = 0;
      } else {
        in[n] = inChar; //хранится как целое
        n++;
      }                    
  }
  if (rComplete) {
    outputs();      
    rComplete = false;
    noCon = false;
  } else { //вкл. режима автоэмуляции через SIM_WAIT c
    if (t_s_ < t) {
        t_ = t_s_ + SIM_WAIT*1000;
        noCon = true;
       }
  }
  if (noCon) { //автоэмуляция
       emul();
       outputs();
  }
  
}

void outputs() {
    //вых. сигнал ШИМ частотных полос,
    //map в диапазон (MIN_PWM_LED... 255)
    for (uint8_t i=0; i<IN_ARR_SIZE-4; i++)
      analogWrite(AI_PIN[i], map(in[i], -127, 127, MIN_PWM_LED, 255));
    //бит-детектор
    if (in[IN_ARR_SIZE-4] == 127) digitalWrite(BEAT0_PIN, HIGH); else digitalWrite(BEAT0_PIN, LOW);
    if (in[IN_ARR_SIZE-3] == 127) digitalWrite(BEAT1_PIN, HIGH); else digitalWrite(BEAT1_PIN, LOW);
    if (in[IN_ARR_SIZE-2] == 127) digitalWrite(BEAT2_PIN, HIGH); else digitalWrite(BEAT2_PIN, LOW);
    //вых сигнал ШИМ ритм-канала, 
    //map в "обратный" диапазон (255... MIN_PWM_OUT)
    analogWrite(RTM_PIN, map(in[IN_ARR_SIZE-1], -127, 127, 255, MIN_PWM_OUT));
    //подсветка на выходе AI_PIN[BCL_PIN_IDX] с уровнем BCL_PWM_OUT
    if (in[0] == -127 && in[1] == -127 && in[2] == -127 && in[3] == -127 &&
      in[4] == -127 && in[5] == -127 && in[6] == -127 && in[7] == -127) {
        analogWrite(RTM_PIN, 0);
        analogWrite(AI_PIN[BCL_PIN_IDX], BCL_PWM_OUT);
      }
}

void emul() {
    if (t_ < t) {
        t_ = t + random(SIM_INT1, SIM_INT2);
        //заполнение входного массива случайными числами
        for (uint8_t i=0; i<IN_ARR_SIZE-4; i++) in[i] = (signed char)random(-127, 127); //каналы ШИМ
        if (random(-127, 127)%2 == 0) in[IN_ARR_SIZE-4] = 0; else in[IN_ARR_SIZE-4] = 127; //бит-каналы
        if (random(-127, 127)%2 == 0) in[IN_ARR_SIZE-3] = 0; else in[IN_ARR_SIZE-3] = 127;
        if (random(-127, 127)%2 == 0) in[IN_ARR_SIZE-2] = 0; else in[IN_ARR_SIZE-2] = 127;
        in[IN_ARR_SIZE-1] = (signed char)random(-127, 127); //ритм-канал
        }
       if (t__ < t) { //сброс сигнала бит-каналов
        t__ = t + SIM_BIT;
        in[IN_ARR_SIZE-4] = 0;
        in[IN_ARR_SIZE-3] = 0;
        in[IN_ARR_SIZE-2] = 0;
       }  
}
